@@ -28,12 +28,17 @@
 #include <libcommute/loperator/es_constructor.hpp>
 #include <libcommute/loperator/space_partition.hpp>
 #include <libcommute/loperator/mapped_basis_view.hpp>
+#include <libcommute/loperator/n_fermion_sector_view.hpp>
 
+#include <algorithm>
 #include <complex>
+#include <iterator>
+#include <memory>
 #include <stdexcept>
 #include <string>
 #include <type_traits>
 #include <utility>
+#include <vector>
 
 using namespace libcommute;
 namespace py = pybind11;
@@ -364,6 +369,15 @@ Bit range spanned by a given elementary space.
 )=",
     py::arg("es")
   )
+  .def("has_algebra",
+    &hs_type::has_algebra,
+R"=(
+Is an elementary space with a given algebra ID found in this Hilbert space?
+
+:param algebra_id: Algebra ID.
+)=",
+   py::arg("algebra_id")
+  )
   .def("algebra_bit_range",
     &hs_type::algebra_bit_range,
 R"=(
@@ -589,6 +603,27 @@ dictionaries ``{(i, j) : O_ij}``.
   );
 }
 
+//
+// Register space_partition::find_connections()
+//
+
+template<typename ScalarType>
+void register_find_connections(py::class_<space_partition> & sp) {
+  sp.def("find_connections",
+  &space_partition::find_connections<hs_type, ScalarType, fermion, boson, spin>,
+R"=(
+Analyze connections between subspaces generated by a given operator
+:math:`\hat O`. The connections are returned as a set of pairs of subspace
+serial numbers, ``(source subspace, destination subspace)``.
+
+:param o: Operator :math:`\hat O`.
+:param hs: Hilbert space used to construct the original space partition.
+:return: A set of ``(source subspace, destination subspace)`` index pairs.
+)=",
+    py::arg("o"), py::arg("hs")
+  );
+}
+
 //
 // Register space_partition
 //
@@ -651,9 +686,29 @@ subspaces of a Hermitian operator.
   register_merge_subspaces<double>(sp);
   register_merge_subspaces<dcomplex>(sp);
 
+  register_find_connections<double>(sp);
+  register_find_connections<dcomplex>(sp);
+
   register_make_space_partition<double>(m);
   register_make_space_partition<dcomplex>(m);
 
+  sp.def("subspace_basis", &space_partition::subspace_basis,
+R"=(
+Build and return a list of indices of all basis states spanning a given
+invariant subspace.
+
+:param index: Index of the invariant subspace.
+)=",
+    py::arg("index")
+  )
+  .def("subspace_bases", &space_partition::subspace_bases,
+R"=(
+Build and return lists of indices of all basis states spanning all invariant
+subspaces in the partition. The returned lists are disjoint and their union
+spans the entire Hilbert space.
+)="
+  );
+
   using f_t = std::function<void(sv_index_type n, sv_index_type sp_index)>;
 
   m.def("foreach",
@@ -784,7 +839,9 @@ corresponding to zero index in each elementary space.
 )=",
     py::arg("O"), py::arg("hs")
   )
-  .def(py::init<std::vector<lop_type<double>> const&, hs_type const&, int>(),
+  .def(py::init<std::vector<lop_type<double>> const&,
+                hs_type const&,
+                unsigned int>(),
 R"=(
 Given a list of operators
 :math:`\{\hat O_1, \hat O_2, \hat O_3, \ldots, \hat O_M\}`, build a mapping
@@ -807,7 +864,9 @@ Mapped values are assigned continuously but without any specific order.
 )=",
     py::arg("O_list"), py::arg("hs"), py::arg("N")
   )
-  .def(py::init<std::vector<lop_type<dcomplex>> const&, hs_type const&, int>(),
+  .def(py::init<std::vector<lop_type<dcomplex>> const&,
+                hs_type const&,
+                unsigned int>(),
 R"=(
 Given a list of operators
 :math:`\{\hat O_1, \hat O_2, \hat O_3, \ldots, \hat O_M\}`, build a mapping
@@ -830,7 +889,7 @@ Mapped values are assigned continuously but without any specific order.
 )=",
     py::arg("O_list"), py::arg("hs"), py::arg("N")
   )
-  .def("__call__", &basis_mapper::make_view_no_ref<py::array_t<double, 0>>,
+  .def("__call__", &basis_mapper::make_view<py::array_t<double, 0>>,
 R"=(
 Make a basis mapping view of a real state vector (1-dimensional NumPy array).
 
@@ -838,7 +897,7 @@ Make a basis mapping view of a real state vector (1-dimensional NumPy array).
 )=",
     py::arg("sv"), py::keep_alive<1, 0>()
   )
-  .def("__call__", &basis_mapper::make_view_no_ref<py::array_t<dcomplex, 0>>,
+  .def("__call__", &basis_mapper::make_view<py::array_t<dcomplex, 0>>,
 R"=(
 Make a basis mapping view of a complex state vector (1-dimensional NumPy array).
 
@@ -864,6 +923,458 @@ Make a basis mapping view of a complex state vector (1-dimensional NumPy array).
 
 ////////////////////////////////////////////////////////////////////////////////
 
+//
+// Register n_fermion_sector_size()
+//
+
+void register_n_fermion_sector_size(py::module_ & m) {
+  m.def(
+    "n_fermion_sector_size",
+     &n_fermion_sector_size<hs_type>,
+     R"=(
+Compute size of a sector (subspace of a full Hilbert space) with a fixed number
+of fermions.
+
+:param hs: Hilbert space.
+:param N: Number of fermions in the sector.
+)=",
+    py::arg("hs"), py::arg("N")
+  );
+}
+
+//
+// Register n_fermion_sector_basis_states()
+//
+
+void register_n_fermion_sector_basis_states(py::module_ & m) {
+  m.def(
+    "n_fermion_sector_basis_states",
+     &n_fermion_sector_basis_states<hs_type>,
+     R"=(
+Return a list of basis states in a sector (subspace of a full Hilbert space)
+with a fixed number of fermions.
+
+:param hs: Hilbert space.
+:param N: Number of fermions in the sector.
+)=",
+    py::arg("hs"), py::arg("N")
+  );
+}
+
+//
+// Register n_fermion_sector_view
+//
+
+template<typename ScalarType>
+void register_n_fermion_sector_view(py::module_ & m,
+                                    std::string const& classname) {
+
+  std::string docstring = "This object is a view of a " +
+                          scalar_type_name<ScalarType>() +
+R"=( state vector (one-dimensional NumPy array) that performs basis state index
+translation from a full Hilbert space to its N-fermion subspace (sector).
+It is accepted by methods of linear operator objects
+:py:func:`LOperatorR.__call__()` and :py:func:`LOperatorC.__call__()`.
+
+:py:class:`)=" + classname + R"=(` can be used in situations where a linear
+operator is known to act only within the N-fermion sector, and it is desirable
+to store vector components only within this particular sector.
+)=";
+
+  py::class_<n_fermion_sector_view<py::array_t<ScalarType, 0>, false>>(
+    m,
+    classname.c_str(),
+    docstring.c_str()
+  )
+  .def(py::init<py::array_t<ScalarType, 0>, hs_type const&, unsigned int>(),
+    R"=(
+Construct an N-fermion sector view of a state vector.
+
+:param sv: The state vector to make the view of.
+:param hs: Hilbert space.
+:param N: Number of fermions in the sector.
+)=",
+    py::arg("sv"), py::arg("hs"), py::arg("N"), py::keep_alive<1, 2>()
+  )
+  .def("map_index",
+       &n_fermion_sector_view<py::array_t<ScalarType, 0>, false>::map_index,
+    R"=(
+Map a basis state index from the full Hilbert space to the sector.
+
+:param index: The basis state index in the full Hilbert space.
+)=",
+    py::arg("index")
+  );
+}
+
+//
+// Register action of linear operators on NFermionSectorView objects
+//
+
+template<typename SrcScalarType, typename DstScalarType, typename LOpType>
+void register_loperator_call_nfsv(py::class_<LOpType> & lop) {
+
+  using scalar_t = typename LOpType::scalar_type;
+  using src_nfsv_t =
+    n_fermion_sector_view<py::array_t<SrcScalarType, 0>, false>;
+  using dst_nfsv_t =
+    n_fermion_sector_view<py::array_t<DstScalarType, 0>, false>;
+
+  std::string src_vector_text = scalar_type_name<SrcScalarType>();
+  std::string dst_vector_text = scalar_type_name<DstScalarType>();
+
+  auto docstring = "\nAct on an N-fermion sector view of a " + src_vector_text +
+    " state vector and write the result through a view of another " +
+    dst_vector_text +
+    " state vector.\n" +
+    R"=(
+:param src: View of the source state vector.
+:param dst: View of the destination state vector.
+)=";
+
+  lop.def("__call__",
+    [](lop_type<scalar_t> const& op, src_nfsv_t const& src, dst_nfsv_t & dst) {
+      op(src, dst);
+    },
+    docstring.c_str(),
+    py::arg("src").noconvert(), py::arg("dst").noconvert()
+  );
+}
+
+////////////////////////////////////////////////////////////////////////////////
+
+using sd_arg_type = std::pair<
+  std::vector<dyn_indices::indices_t>,
+  unsigned int
+>;
+
+std::vector<sector_descriptor<hs_type>> make_sector_descriptors(
+  std::vector<sd_arg_type> const& sectors
+) {
+  std::vector<sector_descriptor<hs_type>> res;
+  res.reserve(sectors.size());
+  for(auto const& sector : sectors) {
+    res.emplace_back(sector_descriptor<hs_type> {
+      std::set<std::tuple<dyn_indices>>(), sector.second
+    });
+    for(auto const& ind : sector.first)
+      res.back().indices.emplace(ind);
+  }
+  return res;
+}
+
+//
+// Register n_fermion_multisector_size()
+//
+
+void register_n_fermion_multisector_size(py::module_ & m) {
+  m.def(
+    "n_fermion_multisector_size",
+    [](hs_type const& hs, std::vector<sd_arg_type> const& sectors) {
+      return n_fermion_multisector_size(hs, make_sector_descriptors(sectors));
+    },
+     R"=(
+Compute size of a multisector (subspace of a full Hilbert space). A multisector
+is a set of all basis states, which have :math:`N_1` particles within a subset
+of fermionic modes :math:`\{S_1\}`, :math:`N_2` particles within another subset
+:math:`\{S_2\}` and so on. There can be any number of individual pairs
+:math:`(\{S_i\}, N_i)` (sectors contributing to the multisector) as long as all
+subsets :math:`\{S_i\}` are disjoint.
+
+:param hs: Hilbert space.
+:param sectors: List of sector descriptors. Each element of the list is a
+                :math:`(\{S_i\}, N_i)` pair, where :math:`\{S_i\}` is a set of
+                index sequences corresponding to the relevant fermionic degrees
+                of freedom.
+)=",
+    py::arg("hs"), py::arg("sectors")
+  );
+}
+
+//
+// Register n_fermion_multisector_basis_states()
+//
+
+void register_n_fermion_multisector_basis_states(py::module_ & m) {
+  m.def(
+    "n_fermion_multisector_basis_states",
+    [](hs_type const& hs, std::vector<sd_arg_type> const& sectors) {
+      return n_fermion_multisector_basis_states(
+        hs,
+        make_sector_descriptors(sectors)
+      );
+    },
+     R"=(
+Return a list of basis states in a multisector (subspace of a full Hilbert
+space). A multisector is a set of all basis states, which have :math:`N_1`
+particles within a subset of fermionic modes :math:`\{S_1\}`, :math:`N_2`
+particles within another subset :math:`\{S_2\}` and so on. There can be any
+number of individual pairs :math:`(\{S_i\}, N_i)` (sectors contributing to the
+multisector) as long as all subsets :math:`\{S_i\}` are disjoint.
+
+:param hs: Hilbert space.
+:param sectors: List of sector descriptors. Each element of the list is a
+                :math:`(\{S_i\}, N_i)` pair, where :math:`\{S_i\}` is a set of
+                index sequences corresponding to the relevant fermionic degrees
+                of freedom.
+)=",
+    py::arg("hs"), py::arg("sectors")
+  );
+}
+
+//
+// Register n_fermion_multisector_view
+//
+
+template<typename ScalarType>
+void register_n_fermion_multisector_view(py::module_ & m,
+                                         std::string const& classname) {
+
+  using view_t = n_fermion_multisector_view<py::array_t<ScalarType, 0>, false>;
+
+  std::string docstring = "This object is a view of a " +
+                          scalar_type_name<ScalarType>() +
+R"=( state vector (one-dimensional NumPy array) that performs basis state index
+translation from a full Hilbert space to an N-fermion multisector.
+It is accepted by methods of linear operator objects
+:py:func:`LOperatorR.__call__()` and :py:func:`LOperatorC.__call__()`.
+
+A multisector is a set of all basis states, which have :math:`N_1`
+particles within a subset of fermionic modes :math:`\{S_1\}`, :math:`N_2`
+particles within another subset :math:`\{S_2\}` and so on. There can be any
+number of individual pairs :math:`(\{S_i\}, N_i)` (sectors contributing to the
+multisector) as long as all subsets :math:`\{S_i\}` are disjoint.
+
+:py:class:`)=" + classname + R"=(` can be used in situations where a linear
+operator is known to act only within the multisector, and it is desirable
+to store vector components only within this particular multisector.
+)=";
+
+  py::class_<view_t>(
+    m,
+    classname.c_str(),
+    docstring.c_str()
+  )
+  .def(py::init([](py::array_t<ScalarType, 0> sv,
+                   hs_type const& hs,
+                   std::vector<sd_arg_type> const& sectors) {
+      return std::make_unique<view_t>(std::move(sv),
+                                      hs,
+                                      make_sector_descriptors(sectors));
+    }),
+    R"=(
+Construct an N-fermion multisector view of a state vector.
+
+:param sv: The state vector to make the view of.
+:param hs: Hilbert space.
+:param sectors: List of sector descriptors. Each element of the list is a
+                :math:`(\{S_i\}, N_i)` pair, where :math:`\{S_i\}` is a set of
+                index sequences corresponding to the relevant fermionic degrees
+                of freedom.
+)=",
+    py::arg("sv"), py::arg("hs"), py::arg("sectors"), py::keep_alive<1, 2>()
+  )
+  .def("map_index",
+    &view_t::map_index,
+    R"=(
+Map a basis state index from the full Hilbert space to the multisector.
+
+:param index: The basis state index in the full Hilbert space.
+)=",
+    py::arg("index")
+  );
+}
+
+//
+// Register action of linear operators on NFermionMultiSectorView objects
+//
+
+template<typename SrcScalarType, typename DstScalarType, typename LOpType>
+void register_loperator_call_nfmsv(py::class_<LOpType> & lop) {
+
+  using scalar_t = typename LOpType::scalar_type;
+  using src_nfmsv_t =
+    n_fermion_multisector_view<py::array_t<SrcScalarType, 0>, false>;
+  using dst_nfmsv_t =
+    n_fermion_multisector_view<py::array_t<DstScalarType, 0>, false>;
+
+  std::string src_vector_text = scalar_type_name<SrcScalarType>();
+  std::string dst_vector_text = scalar_type_name<DstScalarType>();
+
+  auto docstring = "\nAct on an N-fermion multisector view of a " +
+    src_vector_text +
+    " state vector and write the result through a view of another " +
+    dst_vector_text +
+    " state vector.\n" +
+    R"=(
+:param src: View of the source state vector.
+:param dst: View of the destination state vector.
+)=";
+
+  lop.def("__call__",
+    [](lop_type<scalar_t> const& op,
+       src_nfmsv_t const& src,
+       dst_nfmsv_t & dst) {
+      op(src, dst);
+    },
+    docstring.c_str(),
+    py::arg("src").noconvert(), py::arg("dst").noconvert()
+  );
+}
+
+////////////////////////////////////////////////////////////////////////////////
+
+//
+// Define and register make_matrix()
+//
+
+template<typename ScalarType>
+py::array_t<ScalarType> make_matrix(
+  lop_type<ScalarType> const& lop,
+  hs_type const& hs
+) {
+  sv_index_type size = hs.dim();
+
+  auto mat = py::array_t<ScalarType, py::array::f_style>({size, size});
+
+  auto src = py::array_t<ScalarType>(size);
+  auto src_data_ptr = src.mutable_data();
+  std::fill(src_data_ptr, src_data_ptr + size, ScalarType{});
+
+  auto dst = column_view(mat, size);
+
+  for(sv_index_type n = 0; n < size; ++n) {
+    *(src_data_ptr + n) = ScalarType(1);
+    dst.set_column(n);
+    lop(src, dst);
+    *(src_data_ptr + n) = ScalarType(0);
+  }
+
+  return mat;
+}
+
+template<typename ScalarType>
+py::array_t<ScalarType> make_matrix(
+  lop_type<ScalarType> const& lop,
+  std::vector<sv_index_type> const& basis_state_indices
+) {
+  sv_index_type size = basis_state_indices.size();
+
+  auto mat = py::array_t<ScalarType, py::array::f_style>({size, size});
+
+  auto src = py::array_t<ScalarType>(size);
+  auto src_data_ptr = src.mutable_data();
+  std::fill(src_data_ptr, src_data_ptr + size, ScalarType{});
+
+  auto dst = column_view(mat, size);
+
+  auto mapper = basis_mapper(basis_state_indices);
+  auto src_view = mapper.make_const_view(src);
+  auto dst_view = mapper.make_view(dst);
+
+  for(sv_index_type n = 0; n < size; ++n) {
+    *(src_data_ptr + n) = ScalarType(1);
+    dst.set_column(n);
+    lop(src_view, dst_view);
+    *(src_data_ptr + n) = ScalarType(0);
+  }
+
+  return mat;
+}
+
+template<typename ScalarType>
+py::array_t<ScalarType> make_matrix(
+  lop_type<ScalarType> const& lop,
+  std::vector<sv_index_type> const& left_basis_state_indices,
+  std::vector<sv_index_type> const& right_basis_state_indices
+) {
+  sv_index_type size_left = left_basis_state_indices.size();
+  sv_index_type size_right = right_basis_state_indices.size();
+
+  auto mat = py::array_t<ScalarType, py::array::f_style>(
+    {size_left, size_right}
+  );
+
+  auto src = py::array_t<ScalarType>(size_right);
+  auto src_data_ptr = src.mutable_data();
+  std::fill(src_data_ptr, src_data_ptr + size_right, ScalarType{});
+
+  auto dst = column_view(mat, size_left);
+
+  auto mapper_left = basis_mapper(left_basis_state_indices);
+  auto mapper_right = basis_mapper(right_basis_state_indices);
+  auto src_view = mapper_right.make_const_view(src);
+  auto dst_view = mapper_left.make_view(dst);
+
+  for(sv_index_type n = 0; n < size_right; ++n) {
+    *(src_data_ptr + n) = ScalarType(1);
+    dst.set_column(n);
+    lop(src_view, dst_view);
+    *(src_data_ptr + n) = ScalarType(0);
+  }
+
+  return mat;
+}
+
+template<typename ScalarType>
+void register_make_matrix(py::module_ & m) {
+  m.def("make_matrix",
+        [](lop_type<ScalarType> const& lop, hs_type const& hs) ->
+        py::array_t<ScalarType> { return make_matrix(lop, hs); },
+("Make a matrix representation of a given " + scalar_type_name<ScalarType>() +
+R"=( linear operator acting in a Hilbert
+space.
+
+:param lop: Linear operator.
+:param hs: Hilbert space.
+:return: Matrix representation as a two-dimensional NumPy array.
+)=").c_str(),
+    py::arg("lop"), py::arg("hs")
+  )
+  .def("make_matrix",
+       [](lop_type<ScalarType> const& lop,
+          std::vector<sv_index_type> const& basis_state_indices) ->
+       py::array_t<ScalarType> {
+         return make_matrix(lop, basis_state_indices);
+       },
+("Make a matrix representation of a given " + scalar_type_name<ScalarType>() +
+R"=( linear operator acting in a subspace of a Hilbert space spanned by a list
+of basis states.
+
+:param lop: Linear operator.
+:param basis_state_indices: Indices of basis states spanning the subspace.
+:return: Matrix representation as a two-dimensional NumPy array.
+)=").c_str(),
+    py::arg("lop"), py::arg("basis_state_indices")
+  )
+  .def("make_matrix",
+       [](lop_type<ScalarType> const& lop,
+          std::vector<sv_index_type> const& left_basis_state_indices,
+          std::vector<sv_index_type> const& right_basis_state_indices) ->
+       py::array_t<ScalarType> {
+         return make_matrix(lop,
+                            left_basis_state_indices,
+                            right_basis_state_indices);
+       },
+("Make a matrix representation of a given " + scalar_type_name<ScalarType>() +
+R"=( linear operator (mapping) transforming
+states from a subspace of a Hilbert space into states from its other subspace.
+Both subspaces are specified by lists of basis states spanning them.
+
+:param lop: Linear operator.
+:param left_basis_state_indices: Indices of basis states spanning the
+                                 target subspace.
+:param right_basis_state_indices: Indices of basis states spanning the domain.
+:return: Matrix representation as a two-dimensional NumPy array.
+)=").c_str(),
+    py::arg("lop"),
+    py::arg("left_basis_state_indices"),
+    py::arg("right_basis_state_indices")
+  );
+}
+
+////////////////////////////////////////////////////////////////////////////////
+
 //
 // 'loperator' Python module
 //
@@ -872,6 +1383,8 @@ PYBIND11_MODULE(loperator, m) {
 
   m.doc() = "Linear operators in finite-dimensional Hilbert spaces";
 
+  py::module_::import("pycommute.expression");
+
   register_elementary_space(m);
 
   register_elementary_space_fermion(m);
@@ -899,4 +1412,31 @@ PYBIND11_MODULE(loperator, m) {
   register_loperator_call_mbv<dcomplex, dcomplex>(lop_complex);
 
   register_basis_mapper(m);
+
+  register_n_fermion_sector_size(m);
+  register_n_fermion_sector_basis_states(m);
+
+  register_n_fermion_sector_view<double>(m, "NFermionSectorViewR");
+  register_n_fermion_sector_view<dcomplex>(m, "NFermionSectorViewC");
+
+  register_loperator_call_nfsv<double, double>(lop_real);
+  register_loperator_call_nfsv<double, dcomplex>(lop_real);
+  register_loperator_call_nfsv<dcomplex, dcomplex>(lop_real);
+  register_loperator_call_nfsv<double, dcomplex>(lop_complex);
+  register_loperator_call_nfsv<dcomplex, dcomplex>(lop_complex);
+
+  register_n_fermion_multisector_size(m);
+  register_n_fermion_multisector_basis_states(m);
+
+  register_n_fermion_multisector_view<double>(m, "NFermionMultiSectorViewR");
+  register_n_fermion_multisector_view<dcomplex>(m, "NFermionMultiSectorViewC");
+
+  register_loperator_call_nfmsv<double, double>(lop_real);
+  register_loperator_call_nfmsv<double, dcomplex>(lop_real);
+  register_loperator_call_nfmsv<dcomplex, dcomplex>(lop_real);
+  register_loperator_call_nfmsv<double, dcomplex>(lop_complex);
+  register_loperator_call_nfmsv<dcomplex, dcomplex>(lop_complex);
+
+  register_make_matrix<double>(m);
+  register_make_matrix<dcomplex>(m);
 }
